/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package org.openzen.zenscript.formatter;

import org.openzen.zenscript.codemodel.WhitespaceInfo;
import org.openzen.zenscript.codemodel.WhitespacePostComment;
import org.openzen.zenscript.codemodel.expression.Expression;
import org.openzen.zenscript.codemodel.statement.BlockStatement;
import org.openzen.zenscript.codemodel.statement.BreakStatement;
import org.openzen.zenscript.codemodel.statement.CatchClause;
import org.openzen.zenscript.codemodel.statement.ContinueStatement;
import org.openzen.zenscript.codemodel.statement.DoWhileStatement;
import org.openzen.zenscript.codemodel.statement.EmptyStatement;
import org.openzen.zenscript.codemodel.statement.ExpressionStatement;
import org.openzen.zenscript.codemodel.statement.ForeachStatement;
import org.openzen.zenscript.codemodel.statement.IfStatement;
import org.openzen.zenscript.codemodel.statement.LockStatement;
import org.openzen.zenscript.codemodel.statement.ReturnStatement;
import org.openzen.zenscript.codemodel.statement.Statement;
import org.openzen.zenscript.codemodel.statement.StatementVisitor;
import org.openzen.zenscript.codemodel.statement.ThrowStatement;
import org.openzen.zenscript.codemodel.statement.TryCatchStatement;
import org.openzen.zenscript.codemodel.statement.VarStatement;
import org.openzen.zenscript.codemodel.statement.WhileStatement;
import org.openzen.zenscript.codemodel.type.BasicTypeID;

/**
 *
 * @author Hoofdgebruiker
 */
public class StatementFormatter implements StatementVisitor<Void> {
	private final FormattingSettings settings;
	private final StringBuilder result;
	private final ExpressionFormatter expressionFormatter;
	
	private String indent;
	private ParentStatementType position = ParentStatementType.NONE;
	
	public StatementFormatter(String indent, FormattingSettings settings, ExpressionFormatter expressionFormatter) {
		this.indent = indent;
		this.settings = settings;
		result = new StringBuilder();
		this.expressionFormatter = expressionFormatter;
	}
	
	@Override
	public String toString() {
		return result.toString();
	}

	@Override
	public Void visitBlock(BlockStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginBlock(whitespace);
		
		String oldIndent = indent;
		indent = oldIndent + settings.indent;
		for (Statement subStatement : statement.statements) {
			format(ParentStatementType.NONE, subStatement);
		}
		
		WhitespacePostComment postComment = statement.getTag(WhitespacePostComment.class);
		if (postComment != null) {
			writePostComments(postComment.comments);
		}
		
		indent = oldIndent;
		
		endBlock(whitespace);
		return null;
	}

	@Override
	public Void visitBreak(BreakStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("break");
		if (statement.target.label != null)
			result.append(' ').append(statement.target.label);
		result.append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitContinue(ContinueStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("continue");
		if (statement.target.label != null)
			result.append(' ').append(statement.target.label);
		result.append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitDoWhile(DoWhileStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("do");
		if (statement.label != null) {
			if (settings.spaceBeforeLabelColon)
				result.append(' ');
			result.append(':');
			if (settings.spaceAfterLabelColon)
				result.append(' ');
			result.append(statement.label);
		}
		format(ParentStatementType.LOOP, statement.content);
		result.append(" while ");
		appendCondition(statement.condition);
		result.append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitEmpty(EmptyStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append(";\n");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitExpression(ExpressionStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append(statement.expression.accept(expressionFormatter).value)
			  .append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitForeach(ForeachStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("for ");
		for (int i = 0; i < statement.loopVariables.length; i++) {
			if (i > 0)
				result.append(", ");
			
			result.append(statement.loopVariables[i].name);
		}
		result.append(" in ");
		result.append(statement.list.accept(expressionFormatter).value);
		format(ParentStatementType.LOOP, statement.content);
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitIf(IfStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		ParentStatementType position = this.position;
		beginSingleLine(whitespace);
		result.append("if ");
		appendCondition(statement.condition);
		format(statement.onElse == null ? ParentStatementType.IF : ParentStatementType.IF_WITH_ELSE, statement.onThen);
		if (statement.onElse != null) {
			result.append("else");
			format(ParentStatementType.ELSE, statement.onElse);
		}
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitLock(LockStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("lock ");
		result.append(statement.object.accept(expressionFormatter).value);
		statement.content.accept(this);
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitReturn(ReturnStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("return");
		if (statement.value != null) {
			result.append(' ');
			result.append(statement.value.accept(expressionFormatter).value);
		}
		result.append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitThrow(ThrowStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("throw ").append(statement.value.accept(expressionFormatter));
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitTryCatch(TryCatchStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("try");
		if (statement.resource != null) {
			result.append(' ').append(statement.resource.name);
			result.append(" = ");
			result.append(statement.resource.initializer.accept(expressionFormatter).value);
		}
		
		format(ParentStatementType.TRY, statement.content);
		
		for (CatchClause catchClause : statement.catchClauses) {
			result.append(indent).append("catch ");
			if (catchClause.exceptionName != null)
				result.append(catchClause.exceptionName);
			if (catchClause.exceptionType != BasicTypeID.ANY) {
				result.append(" as ");
				catchClause.exceptionType.accept(expressionFormatter.typeFormatter);
			}
			
			format(ParentStatementType.CATCH, catchClause.content);
		}
		if (statement.finallyClause != null) {
			result.append(indent).append("finally ");
			
			format(ParentStatementType.FINALLY, statement.finallyClause);
		}
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitVar(VarStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append(statement.isFinal ? "val " : "var ");
		result.append(statement.name);
		
		if (statement.initializer == null || statement.initializer.type != statement.type) {
			result.append(" as ");
			result.append(statement.type.accept(expressionFormatter.typeFormatter));
		}
		if (statement.initializer != null) {
			result.append(" = ");
			result.append(statement.initializer.accept(expressionFormatter).value);
		}
		result.append(";");
		endSingleLine(whitespace);
		return null;
	}

	@Override
	public Void visitWhile(WhileStatement statement) {
		WhitespaceInfo whitespace = statement.getTag(WhitespaceInfo.class);
		beginSingleLine(whitespace);
		result.append("while");
		if (statement.label != null) {
			if (settings.spaceBeforeLabelColon)
				result.append(' ');
			result.append(':');
			if (settings.spaceAfterLabelColon)
				result.append(' ');
			result.append(statement.label);
		}
		result.append(' ');
		appendCondition(statement.condition);
		
		format(ParentStatementType.LOOP, statement.content);
		endSingleLine(whitespace);
		return null;
	}
	
	private void format(ParentStatementType position, Statement statement) {
		ParentStatementType oldPosition = this.position;
		this.position = position;
		statement.accept(this);
		this.position = oldPosition;
	}
	
	private void appendCondition(Expression condition) {
		if (settings.bracketsAroundConditions)
			result.append('(');
		result.append(condition.accept(expressionFormatter).value);
		if (settings.bracketsAroundConditions)
			result.append(')');
	}
	
	private void beginBlock(WhitespaceInfo whitespace) {
		result.append(settings.getBlockSeparator(indent, position));
		
		if (whitespace != null) {
			if (whitespace.emptyLine) {
				result.append("\n").append(indent);
			}
			writeComments(whitespace.commentsBefore);
		}
	}
	
	private void endBlock(WhitespaceInfo whitespace) {
		if (whitespace != null && !whitespace.commentsAfter.isEmpty())
			result.append(' ').append(whitespace.commentsAfter);
		
		result.append("\n").append(indent).append("}");
		if (position == ParentStatementType.IF_WITH_ELSE) {
			if (settings.elseBracketOnSameLine)
				result.append(" ");
			else
				result.append("\n").append(indent);
		}
	}
	
	private void beginSingleLine(WhitespaceInfo whitespace) {
		result.append(settings.getSingleLineSeparator(indent, position));
		
		if (whitespace != null) {
			if (whitespace.emptyLine) {
				result.append("\n").append(indent);
			}
			writeComments(whitespace.commentsBefore);
		}
	}
	
	private void endSingleLine(WhitespaceInfo whitespace) {
		if (whitespace != null && !whitespace.commentsAfter.isEmpty())
			result.append(' ').append(whitespace.commentsAfter);
		
		if (position == ParentStatementType.IF_WITH_ELSE)
			result.append("\n").append(indent);
	}
	
	private void writeComments(String[] comments) {
		for (String comment : CommentFormatter.format(comments)) {
			result.append(comment).append("\n").append(indent);
		}
	}
	
	private void writePostComments(String[] comments) {
		for (String comment : CommentFormatter.format(comments)) {
			result.append("\n").append(indent).append(comment);
		}
	}
}
